#All abstract methods in the Evolutionary_Optimization module

from abc import abstractmethod
import tpot2
import typing
import tqdm
from tpot2.individual_representations.individual import BaseIndividual
import time
import numpy as np
import copy
import scipy
import os
import pickle
import statistics
from tqdm.dask import TqdmCallback
import distributed
from dask.distributed import Client
from dask.distributed import LocalCluster
from tpot2.selectors import survival_select_NSGA2, tournament_selection_dominated
import math
from tpot2.utils.utils import get_thresholds, beta_interpolation, remove_items, equalize_list

def ind_mutate(ind, rng_):
    rng = np.random.default_rng(rng_)
    return ind.mutate(rng_=rng)

def ind_crossover(ind1, ind2, rng_):
    rng = np.random.default_rng(rng_)
    return ind1.crossover(ind2, rng_=rng)

class BaseEvolver():
    def __init__(   self,
                    individual_generator ,

                    objective_functions,
                    objective_function_weights,
                    objective_names = None,
                    objective_kwargs = None,
                    bigger_is_better = True,

                    population_size = 50,
                    initial_population_size = None,
                    population_scaling = .5,
                    generations_until_end_population = 1,
                    generations = 50,
                    early_stop = None,
                    early_stop_tol = 0.001,


                    max_time_seconds=float("inf"),
                    max_eval_time_seconds=60*5,

                    n_jobs=1,
                    memory_limit="4GB",
                    client=None,

                    survival_percentage = 1,
                    crossover_probability=.2,
                    mutate_probability=.7,
                    mutate_then_crossover_probability=.05,
                    crossover_then_mutate_probability=.05,

                    mutation_functions = [ind_mutate],
                    crossover_functions = [ind_crossover],

                    mutation_function_weights = None,
                    crossover_function_weights = None,

                    n_parents=2,

                    survival_selector = survival_select_NSGA2,
                    parent_selector = tournament_selection_dominated,

                    budget_range = None,
                    budget_scaling = .5,
                    generations_until_end_budget = 1,
                    stepwise_steps = 5,

                    threshold_evaluation_early_stop = None,
                    threshold_evaluation_scaling = .5,
                    min_history_threshold = 20,
                    selection_evaluation_early_stop = None,
                    selection_evaluation_scaling = .5,
                    evaluation_early_stop_steps = None,
                    final_score_strategy = "mean",

                    verbose = 0,
                    periodic_checkpoint_folder = None,
                    callback = None,
                    rng_=None,

                    ) -> None:
        """
        Uses mutation, crossover, and optimization functions to evolve a population of individuals towards the given objective functions.

        Parameters
        ----------
        individual_generator : generator
            Generator that yields new base individuals. Used to generate initial population.
        objective_functions : list of callables
            list of functions that get applied to the individual and return a float or list of floats
            If an objective function returns multiple values, they are all concatenated in order
            with respect to objective_function_weights and early_stop_tol.
        objective_function_weights : list of floats
            list of weights for each objective function. Sign flips whether bigger is better or not
        objective_names : list of strings, default=None
            Names of the objectives. If None, objective0, objective1, etc. will be used
        objective_kwargs : dict, default=None
            Dictionary of keyword arguments to pass to the objective function
        bigger_is_better : bool, default=True
            If True, the objective function is maximized. If False, the objective function is minimized. Use negative weights to reverse the direction.
        population_size : int, default=50
            Size of the population
        initial_population_size : int, default=None
            Size of the initial population. If None, population_size will be used.
        population_scaling : int, default=0.5
            Scaling factor to use when determining how fast we move the threshold moves from the start to end percentile.
        generations_until_end_population : int, default=1
            Number of generations until the population size reaches population_size
        generations : int, default=50
            Number of generations to run
        early_stop : int, default=None
            Number of generations without improvement before early stopping. All objectives must have converged within the tolerance for this to be triggered.
        early_stop_tol : float, list of floats, or None, default=0.001
            -list of floats
                list of tolerances for each objective function. If the difference between the best score and the current score is less than the tolerance, the individual is considered to have converged
                If an index of the list is None, that item will not be used for early stopping
            -int
                If an int is given, it will be used as the tolerance for all objectives
        max_time_seconds : float, default=float("inf")
            Maximum time to run the optimization. If none or inf, will run until the end of the generations.
        max_eval_time_seconds : float, default=60*5
            Maximum time to evaluate a single individual. If none or inf, there will be no time limit per evaluation.
        n_jobs : int, default=1
            Number of processes to run in parallel.
        memory_limit : str, default="4GB"
            Memory limit for each job. See Dask [LocalCluster documentation](https://distributed.dask.org/en/stable/api.html#distributed.Client) for more information.
        client : dask.distributed.Client, default=None
            A dask client to use for parallelization. If not None, this will override the n_jobs and memory_limit parameters. If None, will create a new client with num_workers=n_jobs and memory_limit=memory_limit.
        survival_percentage : float, default=1
            Percentage of the population size to utilize for mutation and crossover at the beginning of the generation. The rest are discarded. Individuals are selected with the selector passed into survival_selector. The value of this parameter must be between 0 and 1, inclusive.
            For example, if the population size is 100 and the survival percentage is .5, 50 individuals will be selected with NSGA2 from the existing population. These will be used for mutation and crossover to generate the next 100 individuals for the next generation. The remainder are discarded from the live population. In the next generation, there will now be the 50 parents + the 100 individuals for a total of 150. Surivival percentage is based of the population size parameter and not the existing population size (current population size when using successive halving). Therefore, in the next generation we will still select 50 individuals from the currently existing 150.
        crossover_probability : float, default=.2
            Probability of generating a new individual by crossover between two individuals.
        mutate_probability : float, default=.7
            Probability of generating a new individual by crossover between one individuals.
        mutate_then_crossover_probability : float, default=.05
            Probability of generating a new individual by mutating two individuals followed by crossover.
        crossover_then_mutate_probability : float, default=.05
            Probability of generating a new individual by crossover between two individuals followed by a mutation of the resulting individual.
        n_parents : int, default=2
            Number of parents to use for crossover. Must be greater than 1.
        survival_selector : function, default=survival_select_NSGA2
            Function to use to select individuals for survival. Must take a matrix of scores and return selected indexes.
            Used to selected population_size * survival_percentage individuals at the start of each generation to use for mutation and crossover.
        parent_selector : function, default=parent_select_NSGA2
            Function to use to select pairs parents for crossover and individuals for mutation. Must take a matrix of scores and return selected indexes.
        budget_range : list [start, end], default=None
            A starting and ending budget to use for the budget scaling.
        budget_scaling float : [0,1], default=0.5
            A scaling factor to use when determining how fast we move the budget from the start to end budget.
        generations_until_end_budget : int, default=1
            The number of generations to run before reaching the max budget.
        stepwise_steps : int, default=1
            The number of staircase steps to take when scaling the budget and population size.
        threshold_evaluation_early_stop : list [start, end], default=None
            starting and ending percentile to use as a threshold for the evaluation early stopping.
            Values between 0 and 100.
        threshold_evaluation_scaling : float [0,inf), default=0.5
            A scaling factor to use when determining how fast we move the threshold moves from the start to end percentile.
            Must be greater than zero. Higher numbers will move the threshold to the end faster.
        min_history_threshold : int, default=0
            The minimum number of previous scores needed before using threshold early stopping.
        selection_evaluation_early_stop : list, default=None
            A lower and upper percent of the population size to select each round of CV.
            Values between 0 and 1.
        selection_evaluation_scaling : float, default=0.5
            A scaling factor to use when determining how fast we move the threshold moves from the start to end percentile.
            Must be greater than zero. Higher numbers will move the threshold to the end faster.
        evaluation_early_stop_steps : int, default=1
            The number of steps that will be taken from the objective function. (e.g., the number of CV folds to evaluate)
        final_score_strategy : str, default="mean"
            The strategy to use when determining the final score for an individual.
            "mean": The mean of all objective scores
            "last": The score returned by the last call. Currently each objective is evaluated with a clone of the individual.
        verbose : int, default=0
            How much information to print during the optimization process. Higher values include the information from lower values.
            0. nothing
            1. progress bar
            2. evaluations progress bar
            3. best individual
            4. warnings
            >=5. full warnings trace
        periodic_checkpoint_folder : str, default=None
            Folder to save the population to periodically. If None, no periodic saving will be done.
            If provided, training will resume from this checkpoint.
        callback : tpot2.CallBackInterface, default=None
            Callback object. Not implemented
        rng_ : Numpy.Random.Generator, None, default=None
            An object for reproducability of experiments. This value will be passed to numpy.random.default_rng() to create an instnce of the genrator to pass to other classes

            - Numpy.Random.Generator
                Will be used to create and lock in Generator instance with 'numpy.random.default_rng()'. Note this will be the same Generator passed in.
            - None
                Will be used to create Generator for 'numpy.random.default_rng()' where a fresh, unpredictable entropy will be pulled from the OS
        """

        self.rng = np.random.default_rng(rng_)

        if threshold_evaluation_early_stop is not None or selection_evaluation_early_stop is not None:
            if evaluation_early_stop_steps is None:
                raise ValueError("evaluation_early_stop_steps must be set when using threshold_evaluation_early_stop or selection_evaluation_early_stop")

        self.individual_generator = individual_generator
        self.population_size = population_size
        self.objective_functions = objective_functions
        self.objective_function_weights = np.array(objective_function_weights)
        self.bigger_is_better = bigger_is_better
        if not bigger_is_better:
            self.objective_function_weights = np.array(self.objective_function_weights)*-1

        self.initial_population_size = initial_population_size
        if self.initial_population_size is None:
            self.cur_population_size = population_size
        else:
            self.cur_population_size = initial_population_size

        self.population_scaling = population_scaling
        self.generations_until_end_population = generations_until_end_population

        self.population_size_list = None


        self.periodic_checkpoint_folder = periodic_checkpoint_folder
        self.verbose  = verbose
        self.callback = callback
        self.generations = generations
        self.n_jobs = n_jobs



        if max_time_seconds is None:
            self.max_time_seconds = float("inf")
        else:
            self.max_time_seconds = max_time_seconds

        #functools requires none for infinite time, doesn't support inf
        if max_eval_time_seconds is not None and math.isinf(max_eval_time_seconds ):
            self.max_eval_time_seconds = None
        else:
            self.max_eval_time_seconds = max_eval_time_seconds




        self.generation = 0


        self.threshold_evaluation_early_stop =threshold_evaluation_early_stop
        self.threshold_evaluation_scaling =  max(0.00001,threshold_evaluation_scaling )
        self.min_history_threshold = min_history_threshold

        self.selection_evaluation_early_stop = selection_evaluation_early_stop
        self.selection_evaluation_scaling =  max(0.00001,selection_evaluation_scaling )
        self.evaluation_early_stop_steps = evaluation_early_stop_steps
        self.final_score_strategy = final_score_strategy

        self.budget_range = budget_range
        self.budget_scaling = budget_scaling
        self.generations_until_end_budget = generations_until_end_budget
        self.stepwise_steps = stepwise_steps

        self.memory_limit = memory_limit

        self.client = client


        self.survival_selector=survival_selector
        self.parent_selector=parent_selector
        self.survival_percentage = survival_percentage

        total_var_p = crossover_probability + mutate_probability + mutate_then_crossover_probability + crossover_then_mutate_probability
        self.crossover_probability = crossover_probability / total_var_p
        self.mutate_probability = mutate_probability  / total_var_p
        self.mutate_then_crossover_probability= mutate_then_crossover_probability / total_var_p
        self.crossover_then_mutate_probability= crossover_then_mutate_probability / total_var_p


        self.mutation_functions = mutation_functions
        self.crossover_functions = crossover_functions

        if mutation_function_weights is None:
            self.mutation_function_weights = [1 for _ in range(len(mutation_functions))]
        else:
            self.mutation_function_weights = mutation_function_weights

        if mutation_function_weights is None:
            self.crossover_function_weights = [1 for _ in range(len(mutation_functions))]
        else:
            self.crossover_function_weights = crossover_function_weights

        self.n_parents = n_parents

        if objective_kwargs is None:
            self.objective_kwargs = {}
        else:
            self.objective_kwargs = objective_kwargs

        # if objective_kwargs is None:
        #     self.objective_kwargs = [{}] * len(self.objective_functions)
        # elif isinstance(objective_kwargs, dict):
        #     self.objective_kwargs = [objective_kwargs] * len(self.objective_functions)
        # else:
        #     self.objective_kwargs = objective_kwargs

        ###########

        if self.initial_population_size != self.population_size:
            self.population_size_list = beta_interpolation(start=self.cur_population_size, end=self.population_size, scale=self.population_scaling, n=generations_until_end_population, n_steps=self.stepwise_steps)
            self.population_size_list = np.round(self.population_size_list).astype(int)

        if self.budget_range is None:
            self.budget_list = None
        else:
            self.budget_list = beta_interpolation(start=self.budget_range[0], end=self.budget_range[1], n=self.generations_until_end_budget, scale=self.budget_scaling, n_steps=self.stepwise_steps)

        if objective_names is None:
            self.objective_names = ["objective"+str(i) for i in range(len(objective_function_weights))]
        else:
            self.objective_names = objective_names

        if self.budget_list is not None:
            if len(self.budget_list) <= self.generation:
                self.budget = self.budget_list[-1]
            else:
                self.budget = self.budget_list[self.generation]
        else:
            self.budget = None


        self.early_stop_tol = early_stop_tol
        self.early_stop = early_stop

        if isinstance(self.early_stop_tol, float):
            self.early_stop_tol = [self.early_stop_tol for _ in range(len(self.objective_names))]

        self.early_stop_tol = [np.inf if tol is None else tol for tol in self.early_stop_tol]

        self.population = None
        self.population_file = None
        if self.periodic_checkpoint_folder is not None:
            self.population_file = os.path.join(self.periodic_checkpoint_folder, "population.pkl")
            if not os.path.exists(self.periodic_checkpoint_folder):
                os.makedirs(self.periodic_checkpoint_folder)
            if os.path.exists(self.population_file):
                self.population = pickle.load(open(self.population_file, "rb"))

                if len(self.population.evaluated_individuals)>0 and "Generation" in self.population.evaluated_individuals.columns:
                    self.generation = self.population.evaluated_individuals['Generation'].max() + 1 #TODO check if this is empty?

        init_names = self.objective_names
        if self.budget_range is not None:
            init_names = init_names + ["Budget"]
        if self.population is None:
            self.population = tpot2.Population(column_names=init_names)
            initial_population = [next(self.individual_generator) for _ in range(self.cur_population_size)]
            self.population.add_to_population(initial_population, self.rng)
            self.population.update_column(self.population.population, column_names="Generation", data=self.generation)


    def optimize(self, generations=None):

        if self.client is not None: #If user passed in a client manually
           self._client = self.client
        else:

            if self.verbose >= 4:
                silence_logs = 30
            elif self.verbose >=5:
                silence_logs = 40
            else:
                silence_logs = 50
            self._cluster = LocalCluster(n_workers=self.n_jobs, #if no client is passed in and no global client exists, create our own
                    threads_per_worker=1,
                    silence_logs=silence_logs,
                    processes=True,
                    memory_limit=self.memory_limit)
            self._client = Client(self._cluster)



        if generations is None:
            generations = self.generations

        start_time = time.time()

        generations_without_improvement = np.array([0 for _ in range(len(self.objective_function_weights))])
        best_scores = [-np.inf for _ in range(len(self.objective_function_weights))]


        self.scheduled_timeout_time = time.time() + self.max_time_seconds


        try:
            #for gen in tnrange(generations,desc="Generation", disable=self.verbose<1):
            done = False
            gen = 0
            if self.verbose >= 1:
                if generations is None or np.isinf(generations):
                    pbar = tqdm.tqdm(total=0)
                else:
                    pbar = tqdm.tqdm(total=generations)
                pbar.set_description("Generation")
            while not done:
                # Generation 0 is the initial population
                if self.generation == 0:
                    if self.population_file is not None:
                        pickle.dump(self.population, open(self.population_file, "wb"))
                    self.evaluate_population()
                    if self.population_file is not None:
                        pickle.dump(self.population, open(self.population_file, "wb"))

                    attempts = 2
                    while len(self.population.population) == 0 and attempts > 0:
                        new_initial_population = [next(self.individual_generator) for _ in range(self.cur_population_size)]
                        self.population.add_to_population(new_initial_population)
                        attempts -= 1
                        self.evaluate_population()

                    if len(self.population.population) == 0:
                        raise Exception("No individuals could be evaluated in the initial population. This may indicate a bug in the configuration, included models, or objective functions. Set verbose>=4 to see the errors that caused individuals to fail.")

                    self.generation += 1
                # Generation 1 is the first generation after the initial population
                else:
                    if time.time() - start_time > self.max_time_seconds:
                        break
                    self.step()

                if self.verbose >= 3:
                    sign = np.sign(self.objective_function_weights)
                    valid_df = self.population.evaluated_individuals[~self.population.evaluated_individuals[self.objective_names].isin(["TIMEOUT","INVALID"]).any(axis=1)][self.objective_names]*sign
                    cur_best_scores = valid_df.max(axis=0)*sign
                    cur_best_scores = cur_best_scores.to_numpy()
                    print("Generation: ", self.generation)
                    for i, obj in enumerate(self.objective_names):
                        print(f"Best {obj} score: {cur_best_scores[i]}")


                if self.early_stop:
                    if self.budget is None or self.budget>=self.budget_range[-1]: #self.budget>=1:
                        #get sign of objective_function_weights
                        sign = np.sign(self.objective_function_weights)
                        #get best score for each objective
                        valid_df = self.population.evaluated_individuals[~self.population.evaluated_individuals[self.objective_names].isin(["TIMEOUT","INVALID"]).any(axis=1)][self.objective_names]*sign
                        cur_best_scores = valid_df.max(axis=0)
                        cur_best_scores = cur_best_scores.to_numpy()
                        #cur_best_scores =  self.population.get_column(self.population.population, column_names=self.objective_names).max(axis=0)*sign #TODO this assumes the current population is the best

                        improved = ( np.array(cur_best_scores) - np.array(best_scores) >= np.array(self.early_stop_tol) )
                        not_improved = np.logical_not(improved)
                        generations_without_improvement = generations_without_improvement * not_improved + not_improved #set to zero if not improved, else increment
                        pass
                        #update best score
                        best_scores = [max(best_scores[i], cur_best_scores[i]) for i in range(len(self.objective_names))]

                        if all(generations_without_improvement>self.early_stop):
                            if self.verbose >= 3:
                                print("Early stop")
                            break

                #save population
                if self.population_file is not None: # and time.time() - last_save_time > 60*10:
                    pickle.dump(self.population, open(self.population_file, "wb"))

                gen += 1
                if self.verbose >= 1:
                    pbar.update(1)

                if generations is not None and gen >= generations:
                    done = True

        except KeyboardInterrupt:
            if self.verbose >= 3:
                print("KeyboardInterrupt")
            self.population.remove_invalid_from_population(column_names=self.objective_names, invalid_value="INVALID")
            self.population.remove_invalid_from_population(column_names=self.objective_names, invalid_value="TIMEOUT")
            self.population.remove_invalid_from_population(column_names="Eval Error", invalid_value="INVALID")
            self.population.remove_invalid_from_population(column_names="Eval Error", invalid_value="TIMEOUT")




        if self.population_file is not None:
            pickle.dump(self.population, open(self.population_file, "wb"))

        if self.client is None: #If we created our own client, close it
            self._client.close()
            self._cluster.close()

        tpot2.utils.get_pareto_frontier(self.population.evaluated_individuals, column_names=self.objective_names, weights=self.objective_function_weights, invalid_values=["TIMEOUT","INVALID"])

    def step(self,):
        if self.population_size_list is not None:
            if self.generation < len(self.population_size_list):
                self.cur_population_size = self.population_size_list[self.generation]
            else:
                self.cur_population_size = self.population_size

        if self.budget_list is not None:
            if len(self.budget_list) <= self.generation:
                self.budget = self.budget_range[-1]
            else:
                self.budget = self.budget_list[self.generation]
        else:
            self.budget = None

        if self.survival_selector is not None:
            n_survivors = max(1,int(self.cur_population_size*self.survival_percentage)) #always keep at least one individual
            self.population.survival_select(    selector=self.survival_selector,
                                                weights=self.objective_function_weights,
                                                columns_names=self.objective_names,
                                                n_survivors=n_survivors,
                                                inplace=True,
                                                rng_=self.rng,)

        self.generate_offspring()
        self.evaluate_population()

        self.generation += 1

    def generate_offspring(self, ): #your EA Algorithm goes here
        parents = self.population.parent_select(selector=self.parent_selector, weights=self.objective_function_weights, columns_names=self.objective_names, k=self.cur_population_size, n_parents=2, rng_=self.rng)
        p = np.array([self.crossover_probability, self.mutate_then_crossover_probability, self.crossover_then_mutate_probability, self.mutate_probability])
        p = p / p.sum()
        var_op_list = self.rng.choice(["crossover", "mutate_then_crossover", "crossover_then_mutate", "mutate"], size=self.cur_population_size, p=p)

        for i, op in enumerate(var_op_list):
            if op == "mutate":
                parents[i] = parents[i][0] #mutations take a single individual

        offspring = self.population.create_offspring2(parents, var_op_list, self.mutation_functions, self.mutation_function_weights, self.crossover_functions, self.crossover_function_weights, add_to_population=True, keep_repeats=False, mutate_until_unique=True, rng_=self.rng)

        self.population.update_column(offspring, column_names="Generation", data=self.generation, )








    # Gets a list of unevaluated individuals in the livepopulation, evaluates them, and removes failed attempts
    # TODO This could probably be an independent function?
    def evaluate_population(self,):

        #Update the sliding scales and thresholds
        # Save population, TODO remove some of these
        if self.population_file is not None: # and time.time() - last_save_time > 60*10:
            pickle.dump(self.population, open(self.population_file, "wb"))
            last_save_time = time.time()


        #Get the current thresholds per step
        self.thresholds = None
        if self.threshold_evaluation_early_stop is not None:
            old_data = self.population.evaluated_individuals[self.objective_names]
            old_data = old_data[old_data[self.objective_names].notnull().all(axis=1)]
            if len(old_data) >= self.min_history_threshold:
                self.thresholds = np.array([get_thresholds(old_data[obj_name],
                                                            start=self.threshold_evaluation_early_stop[0],
                                                            end=self.threshold_evaluation_early_stop[1],
                                                            scale=self.threshold_evaluation_scaling,
                                                            n=self.evaluation_early_stop_steps)
                                        for obj_name in self.objective_names]).T

        #Get the selectors survival rates per step
        if self.selection_evaluation_early_stop is not None:
            lower = self.cur_population_size*self.selection_evaluation_early_stop[0]
            upper = self.cur_population_size*self.selection_evaluation_early_stop[1]
            #survival_counts = self.cur_population_size*(scipy.special.betainc(1,self.selection_evaluation_scaling,np.linspace(0,1,self.evaluation_early_stop_steps))*(upper-lower)+lower)

            survival_counts = np.array(beta_interpolation(start=lower, end=upper, scale=self.selection_evaluation_scaling, n=self.evaluation_early_stop_steps, n_steps=self.evaluation_early_stop_steps))
            self.survival_counts = survival_counts.astype(int)
        else:
            self.survival_counts = None



        if self.evaluation_early_stop_steps is not None:
            if self.survival_counts is None:
                #TODO if we are not using selection method for each step, we can create single threads that run all steps for an individual. No need to come back each step.
                self.evaluate_population_selection_early_stop(survival_counts=self.survival_counts, thresholds=self.thresholds, budget=self.budget)
            else:
                #parallelize one step at a time. After each step, come together and select the next individuals to run the next step on.
                self.evaluate_population_selection_early_stop(survival_counts=self.survival_counts, thresholds=self.thresholds, budget=self.budget)
        else:
            self.evaluate_population_full(budget=self.budget)


        # Save population, TODO remove some of these
        if self.population_file is not None: # and time.time() - last_save_time > 60*10:
            pickle.dump(self.population, open(self.population_file, "wb"))
            last_save_time = time.time()

    def evaluate_population_full(self, budget=None):
        individuals_to_evaluate = self.get_unevaluated_individuals(self.objective_names, budget=budget,)

        #print("evaluating this many individuals: ", len(individuals_to_evaluate))

        if len(individuals_to_evaluate) == 0:
            if self.verbose > 3:
                print("No new individuals to evaluate")
            return

        if self.max_eval_time_seconds is not None:
            theoretical_timeout = self.max_eval_time_seconds * math.ceil(len(individuals_to_evaluate) / self.n_jobs)
            theoretical_timeout = theoretical_timeout*2
        else:
            theoretical_timeout = np.inf
        scheduled_timeout_time_left = self.scheduled_timeout_time - time.time()
        parallel_timeout = min(theoretical_timeout, scheduled_timeout_time_left)
        if parallel_timeout < 0:
            parallel_timeout = 10

        #scores = tpot2.utils.eval_utils.parallel_eval_objective_list(individuals_to_evaluate, self.objective_functions, self.n_jobs, verbose=self.verbose, timeout=self.max_eval_time_seconds, budget=budget, n_expected_columns=len(self.objective_names), client=self._client, parallel_timeout=parallel_timeout, **self.objective_kwargs)
        scores, start_times, end_times, eval_errors = tpot2.utils.eval_utils.parallel_eval_objective_list2(individuals_to_evaluate, self.objective_functions, verbose=self.verbose, max_eval_time_seconds=self.max_eval_time_seconds, budget=budget, n_expected_columns=len(self.objective_names), client=self._client, **self.objective_kwargs)

        self.population.update_column(individuals_to_evaluate, column_names=self.objective_names, data=scores)
        if budget is not None:
            self.population.update_column(individuals_to_evaluate, column_names="Budget", data=budget)

        self.population.update_column(individuals_to_evaluate, column_names="Submitted Timestamp", data=start_times)
        self.population.update_column(individuals_to_evaluate, column_names="Completed Timestamp", data=end_times)
        self.population.update_column(individuals_to_evaluate, column_names="Eval Error", data=eval_errors)
        self.population.remove_invalid_from_population(column_names="Eval Error")
        self.population.remove_invalid_from_population(column_names="Eval Error", invalid_value="TIMEOUT")

    def get_unevaluated_individuals(self, column_names, budget=None, individual_list=None):
        if individual_list is not None:
            cur_pop = np.array(individual_list)
        else:
            cur_pop = np.array(self.population.population)

        if all([name_step in self.population.evaluated_individuals.columns for name_step in column_names]):
            if budget is not None:
                offspring_scores = self.population.get_column(cur_pop, column_names=column_names+["Budget"], to_numpy=False)
                #Individuals are unevaluated if we have a higher budget OR if any of the objectives are nan
                unevaluated_filter = lambda i: any(offspring_scores.loc[offspring_scores.index[i]][column_names].isna()) or (offspring_scores.loc[offspring_scores.index[i]]["Budget"] < budget)
            else:
                offspring_scores = self.population.get_column(cur_pop, column_names=column_names, to_numpy=False)
                unevaluated_filter = lambda i: any(offspring_scores.loc[offspring_scores.index[i]][column_names].isna())
            unevaluated_individuals_this_step = [i for i in range(len(cur_pop)) if unevaluated_filter(i)]
            return cur_pop[unevaluated_individuals_this_step]

        else: #if column names are not in the evaluated_individuals, then we have not evaluated any individuals yet
            for name_step in column_names:
                self.population.evaluated_individuals[name_step] = np.nan
            return cur_pop

    def evaluate_population_selection_early_stop(self,survival_counts, thresholds=None, budget=None):


        survival_selector = tpot2.selectors.survival_select_NSGA2

        ################

        objective_function_signs = np.sign(self.objective_function_weights)


        cur_individuals = self.population.population.copy()

        all_step_names = []
        for step in range(self.evaluation_early_stop_steps):
            if budget is None:
                this_step_names = [f"{n}_step_{step}" for n in self.objective_names]
            else:
                this_step_names = [f"{n}_budget_{budget}_step_{step}" for n in self.objective_names]

            all_step_names.append(this_step_names)

            unevaluated_individuals_this_step = self.get_unevaluated_individuals(this_step_names, budget=None, individual_list=cur_individuals)

            if len(unevaluated_individuals_this_step) == 0:
                if self.verbose > 3:
                    print("No new individuals to evaluate")
                continue

            if self.max_eval_time_seconds is not None:
                theoretical_timeout = self.max_eval_time_seconds * math.ceil(len(unevaluated_individuals_this_step) / self.n_jobs)
                theoretical_timeout = theoretical_timeout*2
            else:
                theoretical_timeout = np.inf
            scheduled_timeout_time_left = self.scheduled_timeout_time - time.time()
            parallel_timeout = min(theoretical_timeout, scheduled_timeout_time_left)
            if parallel_timeout < 0:
                parallel_timeout = 10

            scores, start_times, end_times, eval_errors = tpot2.utils.eval_utils.parallel_eval_objective_list2(individual_list=unevaluated_individuals_this_step,
                                    objective_list=self.objective_functions,
                                    verbose=self.verbose,
                                    max_eval_time_seconds=self.max_eval_time_seconds,
                                    step=step,
                                    budget = self.budget,
                                    generation = self.generation,
                                    n_expected_columns=len(self.objective_names),
                                    client=self._client,
                                    **self.objective_kwargs,
                                    )
            
            self.population.update_column(unevaluated_individuals_this_step, column_names=this_step_names, data=scores)
            self.population.update_column(unevaluated_individuals_this_step, column_names="Submitted Timestamp", data=start_times)
            self.population.update_column(unevaluated_individuals_this_step, column_names="Completed Timestamp", data=end_times)
            self.population.update_column(unevaluated_individuals_this_step, column_names="Eval Error", data=eval_errors)

            self.population.remove_invalid_from_population(column_names="Eval Error")
            self.population.remove_invalid_from_population(column_names="Eval Error", invalid_value="TIMEOUT")

            #remove invalids:
            invalids = []
            #find indeces of invalids

            for j in range(len(scores)):
                if  any([s=="INVALID" for s in scores[j]]):
                    invalids.append(j)

            for j in range(len(scores)):
                if  any([s=="TIMEOUT" for s in scores[j]]):
                    invalids.append(j)


            #already evaluated
            already_evaluated = list(set(cur_individuals) - set(unevaluated_individuals_this_step))
            #evaluated and valid
            valid_evaluations_this_step = remove_items(unevaluated_individuals_this_step,invalids)
            #update cur_individuals with current individuals with valid scores
            cur_individuals = np.concatenate([already_evaluated, valid_evaluations_this_step])


            #Get average scores

            #array of shape (steps, individuals, objectives)
            offspring_scores = [self.population.get_column(cur_individuals, column_names=step_names) for step_names in all_step_names]
            offspring_scores = np.array(offspring_scores)
            if self.final_score_strategy == 'mean':
                offspring_scores  = offspring_scores.mean(axis=0)
            elif self.final_score_strategy == 'last':
                offspring_scores = offspring_scores[-1]


            #if last step, add the final metrics
            if step == self.evaluation_early_stop_steps-1:
                self.population.update_column(cur_individuals, column_names=self.objective_names, data=offspring_scores)
                if budget is not None:
                    self.population.update_column(cur_individuals, column_names="Budget", data=budget)
                return

            #If we have more threads than remaining individuals, we may as well evaluate the extras too
            if self.n_jobs < len(cur_individuals):
                #Remove based on thresholds
                if thresholds is not None:
                    threshold = thresholds[step]
                    invalids = []
                    for i in range(len(offspring_scores)):

                        if all([s*w>t*w for s,t,w in zip(offspring_scores[i],threshold,objective_function_signs)  ]):
                            invalids.append(i)

                    if len(invalids) > 0:

                        max_to_remove = min(len(cur_individuals) - self.n_jobs, len(invalids))

                        if max_to_remove < len(invalids):
                            invalids = np.random.choice(invalids, max_to_remove, replace=False)

                        cur_individuals = remove_items(cur_individuals,invalids)
                        offspring_scores = remove_items(offspring_scores,invalids)

                #Remove based on selection
                if survival_counts is not None:
                    if step < self.evaluation_early_stop_steps - 1 and survival_counts[step]>1: #don't do selection for the last loop since they are completed
                        k = survival_counts[step] + len(invalids) #TODO can remove the min if the selections method can ignore k>population size
                        if len(cur_individuals)> 1 and k > self.n_jobs and k < len(cur_individuals):
                            weighted_scores = np.array([s * self.objective_function_weights for s in offspring_scores ])

                            new_population_index = survival_selector(weighted_scores, k=k)
                            cur_individuals = np.array(cur_individuals)[new_population_index]



import numpy as np

from.nsgaii import nondominated_sorting, crowding_distance, dominates

#based on deap
def tournament_selection_dominated(scores, k, rng_=None, n_parents=2):
    """Select the best individual among *tournsize* randomly chosen
    individuals, *k* times. The returned list contains the indices of the chosen *individuals*.
    :param scores: The score matrix, where rows the individulas and the columns are the corresponds to scores on different objectives.
    :param k: The number of individuals to select.
    :param tournsize: The number of individuals participating in each tournament.
    :param score_index: The number of individuals participating in each tournament.
    :returns: A list of indices of selected individuals.
    This function uses the :func:`~random.choice` function from the python base
    :mod:`random` module.
    """

    rng = np.random.default_rng(rng_)
    pareto_fronts = nondominated_sorting(scores)

    # chosen = list(itertools.chain.from_iterable(fronts))
    # if len(chosen) >= k:
    #     return chosen[0:k]

    crowding_dict = {}
    chosen = []
    current_front_number = 0
    while current_front_number < len(pareto_fronts):

        current_front = np.array(list(pareto_fronts[current_front_number]))
        front_scores = [scores[i] for i in current_front]
        crowding_distances = crowding_distance(front_scores)
        for i, crowding in zip(current_front,crowding_distances):
            crowding_dict[i] = crowding

        current_front_number += 1


    chosen = []
    for i in range(k*n_parents):
        asp1 = rng.choice(len(scores))
        asp2 = rng.choice(len(scores))

        if dominates(scores[asp1], scores[asp2]):
            chosen.append(asp1)
        elif dominates(scores[asp2], scores[asp1]):
            chosen.append(asp2)

        elif crowding_dict[asp1] > crowding_dict[asp2]:
            chosen.append(asp1)
        elif crowding_dict[asp1] < crowding_dict[asp2]:
            chosen.append(asp2)

        else:
            chosen.append(rng.choice([asp1,asp2]))

    return np.reshape(chosen, (k, n_parents))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#TODO update this
from setuptools import setup, find_packages


def calculate_version():
    initpy = open('tpot2/_version.py').read().split('\n')
    version = list(filter(lambda x: '__version__' in x, initpy))[0].split('\'')[1]
    return version


package_version = calculate_version()

setup(
    name='TPOT2',
    python_requires='<3.12', #for configspace compatibility
    version=package_version,
    author='Pedro Ribeiro',
    packages=find_packages(),
    url='https://github.com/EpistasisLab/tpot2',
    license='GNU/LGPLv3', #TODO
    entry_points={'console_scripts': ['tpot2=tpot2:main', ]},
    description=('Tree-based Pipeline Optimization Tool'),
    long_description='''
A Python tool that automatically creates and optimizes machine learning pipelines using genetic programming.


''',
    zip_safe=True,
    install_requires=['numpy>=1.26.4',
                      'scipy>=1.3.1',
                      'scikit-learn>=1.3.0',
                      'update_checker>=0.16',
                      'tqdm>=4.36.1',
                      'stopit>=1.1.1',
                      'pandas>=2.2.0',
                      'joblib>=1.1.1',
                      'xgboost>=1.7.0',
                      'matplotlib>=3.6.2',
                      'traitlets>=5.8.0',
                      'lightgbm>=3.3.3',
                      'optuna>=3.0.5',
                      'baikal>=0.4.2',
                      'networkx>=3.0',
                      'dask>=2024.4.2',
                      'distributed>=2024.4.2',
                      'dask-expr>=1.0.12',
                      'dask-jobqueue>=0.8.5',
                      'func_timeout>=4.3.5',
                      'configspace>=0.7.1',
                     ],
    extras_require={
        'skrebate': ['skrebate>=0.3.4'],
        'mdr': ['scikit-mdr>=0.4.4'],
        'sklearnex' : ['scikit-learn-intelex>=2023.2.1']
    },
    classifiers=[
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)',
        'Programming Language :: Python :: 3.10',
        'Topic :: Scientific/Engineering :: Artificial Intelligence'
    ],
    keywords=['pipeline optimization', 'hyperparameter optimization', 'data science', 'machine learning', 'genetic programming', 'evolutionary computation'],
)



# %%
# -*- coding: utf-8 -*-

"""Copyright (c) 2015 The auto-sklearn developers. All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

  a. Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.
  b. Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
  c. Neither the name of the auto-sklearn Developers  nor the names of
     its contributors may be used to endorse or promote products
     derived from this software without specific prior written
     permission.


THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.
"""
#TODO support np arrays

import numpy as np
from scipy import sparse

from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.utils import check_array
from sklearn.preprocessing import OneHotEncoder
import sklearn
import sklearn.impute

import pandas as pd
from pandas.api.types import is_numeric_dtype
import sklearn.compose


class ColumnSimpleImputer(BaseEstimator, TransformerMixin):
    def __init__(self,  columns="all",         
                        missing_values=np.nan,
                        strategy="mean",
                        fill_value=None,
                        copy=True,
                        add_indicator=False,
                        keep_empty_features=False,):
        
        self.columns = columns
        self.missing_values = missing_values
        self.strategy = strategy
        self.fill_value = fill_value
        self.copy = copy
        self.add_indicator = add_indicator
        self.keep_empty_features = keep_empty_features


    def fit(self, X, y=None):
        """Fit OneHotEncoder to X, then transform X.

        Equivalent to self.fit(X).transform(X), but more convenient and more
        efficient. See fit for the parameters, transform for the return value.

        Parameters
        ----------
        X : array-like or sparse matrix, shape=(n_samples, n_features)
            Dense array or sparse matrix.
        y: array-like {n_samples,} (Optional, ignored)
            Feature labels
        """

        if (self.columns == "categorical" or self.columns == "numeric") and not isinstance(X, pd.DataFrame):
            raise ValueError(f"Invalid value for columns: {self.columns}. "
                             "Only 'all' or <list> is supported for np arrays")

        if self.columns == "categorical":
            self.columns_ = list(X.select_dtypes(exclude='number').columns)
        elif self.columns == "numeric":
            self.columns_ =  [col for col in X.columns if is_numeric_dtype(X[col])]
        elif self.columns == "all":
            if isinstance(X, pd.DataFrame):
                self.columns_ = X.columns
            else:
                self.columns_ = list(range(X.shape[1]))
        elif isinstance(self.columns, list):
            self.columns_ = self.columns
        else:
            raise ValueError(f"Invalid value for columns: {self.columns}")
        
        if len(self.columns_) == 0:
            return self
        
        self.imputer = sklearn.impute.SimpleImputer(missing_values=self.missing_values,
                                                    strategy=self.strategy,
                                                    fill_value=self.fill_value,
                                                    copy=self.copy,
                                                    add_indicator=self.add_indicator,
                                                    keep_empty_features=self.keep_empty_features)
        
        if isinstance(X, pd.DataFrame):
            self.imputer.set_output(transform="pandas")

        if isinstance(X, pd.DataFrame):
            self.imputer.fit(X[self.columns_], y)
        else:
            self.imputer.fit(X[:, self.columns_], y)

        return self

    def transform(self, X):
        """Transform X using one-hot encoding.

        Parameters
        ----------
        X : array-like or sparse matrix, shape=(n_samples, n_features)
            Dense array or sparse matrix.

        Returns
        -------
        X_out : sparse matrix if sparse=True else a 2-d array, dtype=int
            Transformed input.
        """
        if len(self.columns_) == 0:
            return X

        if isinstance(X, pd.DataFrame):
            X = X.copy()
            X[self.columns_] = self.imputer.transform(X[self.columns_])
            return X
        else:
            X = np.copy(X)
            X[:, self.columns_] = self.imputer.transform(X[:, self.columns_])
            return X





import numpy as np
from tpot2 import config
import networkx as nx
from abc import abstractmethod
import matplotlib.pyplot as plt
import sklearn
import tpot2
import sklearn.pipeline
from typing import Generator
import optuna
from itertools import combinations
from .graph_utils import graph_utils
import itertools
import baikal
import copy
from .. import BaseIndividual

class NodeLabel():
    def __init__(self, *,

        #intialized, but may change later
            method_class = None, #transformer or baseestimator
            hyperparameters=None,
            label=None,
    ):

        #intializable, but may change later
        self.method_class = method_class #transformer or baseestimator
        self.hyperparameters = hyperparameters
        self.label = label
        self._params = None



from functools import partial
#@https://stackoverflow.com/questions/20530455/isomorphic-comparison-of-networkx-graph-objects-instead-of-the-default-address

class GraphKey():
    '''
    A class that can be used as a key for a graph.

    Parameters
    ----------
    graph : (nx.Graph)
        The graph to use as a key. Node Attributes are used for the hash.
    matched_label : (str)
        The node attribute to consider for the hash.
    '''

    def __init__(self, graph, matched_label='label') -> None:#['hyperparameters', 'method_class']) -> None:


        self.graph = graph
        self.matched_label = matched_label
        self.node_match = partial(node_match, matched_labels=[matched_label])
        self.key = int(nx.weisfeiler_lehman_graph_hash(self.graph, node_attr=self.matched_label),16) #hash(tuple(sorted([val for (node, val) in self.graph.degree()])))


    #If hash is different, node is definitely different
    # https://arxiv.org/pdf/2002.06653.pdf
    def __hash__(self) -> int:

        return self.key

    #If hash is same, use __eq__ to know if they are actually different
    def __eq__(self, other):
        return nx.is_isomorphic(self.graph, other.graph, node_match=self.node_match)

def node_match(n1,n2, matched_labels):
    return all( [ n1[m] == n2[m] for m in matched_labels])


class GraphIndividual(BaseIndividual):
    '''
    An individual that contains a template for a graph sklearn pipeline.

    Parameters
    ----------
    root_config_dict : {dict with format {method class: param_function}}
        A dictionary of methods and functions that return a dictionary of hyperparameters.
        Used to create the root node of the graph.
    inner_config_dict : {dict with format {method class: param_function}}
        A dictionary of methods and functions that return a dictionary of hyperparameters.
        Used to create the inner nodes of the graph. If None, uses root_config_dict.
    leaf_config_dict : {dict with format {method class: param_function}}
        A dictionary of methods and functions that return a dictionary of hyperparameters.
        Used to create the leaf nodes of the graph. If not None, then all leafs must be created from this dictionary.
        Otherwise leaves will be created from inner_config_dict.
    initial_graph : (nx.DiGraph or list):
        A graph to initialize the individual with.
        If a list, it will initialize a linear graph with the methods in the list in the sequence provided.
        If the items in the list are dictionaries, nodes will be itialized with those dictionaries.
        Strings in the list correspond to the default configuration files. They can be 'Selector', 'Regressor', 'Transformer', 'Classifier'.
    max_depth : (int)
        The maximum depth of the graph as measured by the shortest distance from the root.
    max_size : (int)
        The maximum number of nodes in the graph.
    max_children : (int)
        The maximum number of children a node can have.
    name : (str)
        The name of the individual.
    crossover_same_depth : (bool)
        If true, then crossover will only occur between nodes of the same depth as measured by the shortest distance from the root.
    crossover_same_recursive_depth : (bool)
        If the graph is recursive, then crossover will only occur between graphs of the same recursive depth as measured by the shortest distance from the root.
    '''
    def __init__(
                self,
                root_config_dict,
                inner_config_dict=None,
                leaf_config_dict=None,
                initial_graph = None,
                max_size = np.inf,
                linear_pipeline = False,
                name=None,
                crossover_same_depth = False,
                crossover_same_recursive_depth = True,

                hyperparameter_probability = 1,
                hyper_node_probability = 0,
                hyperparameter_alpha = 1,

                unique_subset_values = None,
                initial_subset_values = None,
                rng_=None,
                ):

        self.__debug = False

        rng = np.random.default_rng(rng_)

        self.root_config_dict = root_config_dict
        self.inner_config_dict = inner_config_dict
        self.leaf_config_dict = leaf_config_dict


        self.max_size = max_size
        self.name = name

        self.crossover_same_depth = crossover_same_depth
        self.crossover_same_recursive_depth = crossover_same_recursive_depth

        self.unique_subset_values = unique_subset_values
        self.initial_subset_values = initial_subset_values

        self.hyperparameter_probability = hyperparameter_probability
        self.hyper_node_probability = hyper_node_probability
        self.hyperparameter_alpha = hyperparameter_alpha

        if self.unique_subset_values is not None:
            self.row_subset_selector = tpot2.representations.SubsetSelector(rng_=rng, values=unique_subset_values, initial_set=initial_subset_values,k=20)

        if isinstance(initial_graph, nx.DiGraph):
            self.graph = initial_graph
            self.root = list(nx.topological_sort(self.graph))[0]

            if self.leaf_config_dict is not None and len(self.graph.nodes) == 1:
                first_leaf = create_node(self.leaf_config_dict, rng_=rng)
                self.graph.add_edge(self.root,first_leaf)

        elif isinstance(initial_graph, list):
            node_list = []
            for item in initial_graph:
                if isinstance(item, dict):
                    node_list.append(create_node(item, rng_=rng))
                elif isinstance(item, str):
                    if item == 'Selector':
                            from tpot2.config import selector_config_dictionary
                            node_list.append(create_node(selector_config_dictionary, rng_=rng))
                    elif  item == 'Regressor':
                            from tpot2.config import regressor_config_dictionary
                            node_list.append(create_node(regressor_config_dictionary, rng_=rng))
                    elif  item == 'Transformer':
                            from tpot2.config import transformer_config_dictionary
                            node_list.append(create_node(transformer_config_dictionary, rng_=rng))
                    elif  item == 'Classifier':
                            from tpot2.config import classifier_config_dictionary
                            node_list.append(create_node(classifier_config_dictionary, rng_=rng))

            self.graph = nx.DiGraph()
            for child, parent in zip(node_list, node_list[1:]):
                self.graph.add_edge(parent, child)

            self.root = node_list[-1]

        else:
            self.graph = nx.DiGraph()

            self.root = create_node(self.root_config_dict, rng_=rng)
            self.graph.add_node(self.root)

            if self.leaf_config_dict is not None:
                first_leaf = create_node(self.leaf_config_dict, rng_=rng)
                self.graph.add_edge(self.root,first_leaf)



        self.initialize_all_nodes(rng_=rng)

        #self.root =list(nx.topological_sort(self.graph))[0]


        self.mutate_methods_list =     [self._mutate_hyperparameters,
                                        self._mutate_replace_node,
                                        self._mutate_remove_node,
                                        ]

        self.crossover_methods_list = [
                                        self._crossover_swap_branch,
                                        ]


        if self.inner_config_dict is not None:
            self.mutate_methods_list.append(self._mutate_insert_inner_node)
            self.crossover_methods_list.append(self._crossover_take_branch) #this is the only crossover method that can create inner nodes
            if not linear_pipeline:
                self.mutate_methods_list.append(self._mutate_insert_bypass_node)
                self.mutate_methods_list.append(self._mutate_remove_edge)
                self.mutate_methods_list.append(self._mutate_add_edge)

        if not linear_pipeline and (self.leaf_config_dict is not None or self.inner_config_dict is not None):
            self.mutate_methods_list.append(self._mutate_insert_leaf)




        if self.unique_subset_values is not None:
            self.crossover_methods_list.append(self._crossover_row_subsets)
            self.mutate_methods_list.append(self._mutate_row_subsets )

        self.optimize_methods_list = [ #self._optimize_optuna_single_method_full_pipeline,
                                        self._optimize_optuna_all_methods_full_pipeline]

        self.key = None

    def select_config_dict(self, node):
        #check if the node is root, leaf, or inner
        if len(list(self.graph.predecessors(node))) == 0: #root
            return self.root_config_dict
        elif self.leaf_config_dict is not None and len(list(self.graph.successors(node))) == 0: #leaf
            return self.leaf_config_dict
        else: #inner
            return self.inner_config_dict


    def initialize_all_nodes(self, rng_=None):
        rng = np.random.default_rng(rng_)
        for node in self.graph:
            if isinstance(node,GraphIndividual):
                continue
            if node.method_class is None:
                node.method_class = rng.choice(list(self.select_config_dict(node).keys()))
            if node.hyperparameters is None:
                get_hyperparameter(self.select_config_dict(node)[node.method_class], nodelabel=node,  alpha=self.hyperparameter_alpha, hyperparameter_probability=self.hyperparameter_probability)


    def fix_noncompliant_leafs(self, rng_=None):
        rng = np.random.default_rng(rng_)
        leafs = [node for node in self.graph.nodes if len(list(self.graph.successors(node)))==0]
        compliant_leafs = []
        noncompliant_leafs = []
        for leaf in leafs:
            if leaf.method_class in self.leaf_config_dict:
                compliant_leafs.append(leaf)
            else:
                noncompliant_leafs.append(leaf)

        #find all good leafs. If no good leaves exist, create a new one
        if len(compliant_leafs) == 0:
            first_leaf = NodeLabel(config_dict=self.leaf_config_dict)
            first_leaf.method_class = rng.choice(list(first_leaf.config_dict.keys())) #TODO: check when there is no new method
            first_leaf.hyperparameters = first_leaf.config_dict[first_leaf.method_class](config.hyperparametersuggestor)
            get_hyperparameter(self.select_config_dict(first_leaf)[first_leaf.method_class], nodelabel=first_leaf,  alpha=self.hyperparameter_alpha, hyperparameter_probability=self.hyperparameter_probability)
            compliant_leafs.append(first_leaf)

        #connect bad leaves to good leaves (making them internal nodes)
        if len(noncompliant_leafs) > 0:
            for node in noncompliant_leafs:
                self.graph.add_edge(node, rng.choice(compliant_leafs))




    def _merge_duplicated_nodes(self):

        graph_changed = False
        merged = False
        while(not merged):
            node_list = list(self.graph.nodes)
            merged = True
            for node, other_node in itertools.product(node_list, node_list):
                if node is other_node or isinstance(node,GraphIndividual) or isinstance(other_node,GraphIndividual):
                    continue

                #If nodes are same class/hyperparameters
                if node.method_class == other_node.method_class and node.hyperparameters == other_node.hyperparameters:
                    node_children = set(self.graph.successors(node))
                    other_node_children = set(self.graph.successors(other_node))
                    #if nodes have identical children, they can be merged
                    if node_children == other_node_children:
                        for other_node_parent in list(self.graph.predecessors(other_node)):
                            if other_node_parent not in self.graph.predecessors(node):
                                self.graph.add_edge(other_node_parent,node)

                        self.graph.remove_node(other_node)
                        merged=False
                        graph_changed = True
                        break

        return graph_changed

    #returns a flattened pipeline
    def flatten_pipeline(self,depth=0):
        flattened_full_graph = self.graph.copy()
        remove_list = []
        for node in flattened_full_graph:
            if isinstance(node,GraphIndividual):
                flattened = node.flatten_pipeline(depth+1)
                roots = graph_utils.get_roots(flattened)
                leaves = graph_utils.get_leaves(flattened)

                n1_s = flattened_full_graph.successors(node)
                n1_p = flattened_full_graph.predecessors(node)

                remove_list.append(node)


                flattened_full_graph = nx.compose(flattened_full_graph, flattened)


                flattened_full_graph.add_edges_from([ (n2, n) for n in n1_s for n2 in leaves])
                flattened_full_graph.add_edges_from([ (n, n2) for n in n1_p for n2 in roots])
            else:
                flattened_full_graph.nodes[node]['recursive depth'] = depth


        for node in remove_list:
            flattened_full_graph.remove_node(node)

        if self.unique_subset_values is not None:
            for node in flattened_full_graph:
                if "subset_values" not in flattened_full_graph.nodes[node]:
                    flattened_full_graph.nodes[node]["subset_values"] = list(self.row_subset_selector.subsets)
                else:
                    #intersection
                    flattened_full_graph.nodes[node]["subset_values"] = list(set(flattened_full_graph.nodes[node]["subset_values"]) & set(self.row_subset_selector.subsets))

        return flattened_full_graph

    def get_num_nodes(self,):
        num_nodes = 0

        for node in self.graph.nodes:
            if isinstance(node, GraphIndividual):
                num_nodes+= node.get_num_nodes()
            else:
                num_nodes += 1

        return num_nodes


    def export_nested_pipeline(self, **graph_pipeline_args):

        flattened_full_graph = self.graph.copy()
        remove_list = []
        for node in list(flattened_full_graph.nodes):
            if isinstance(node,GraphIndividual):
                gp = node.export_pipeline(**graph_pipeline_args)

                n1_s = flattened_full_graph.successors(node)
                n1_p = flattened_full_graph.predecessors(node)

                remove_list.append(node)

                flattened_full_graph.add_node(gp)


                flattened_full_graph.add_edges_from([ (gp, n) for n in n1_s])
                flattened_full_graph.add_edges_from([ (n, gp) for n in n1_p])


        for node in remove_list:
            flattened_full_graph.remove_node(node)

        estimator_graph = flattened_full_graph

        #mapping = {node:node.method_class(**node.hyperparameters) for node in estimator_graph}
        label_remapping = {}
        label_to_instance = {}

        for node in estimator_graph:
            found_unique_label = False
            i=1
            while not found_unique_label:
                print(type(node))
                if type(node) is tpot2.GraphPipeline:
                    label = "GraphPipeline_{0}".format( i)
                else:
                    label = "{0}_{1}".format(node.method_class.__name__, i)
                if label not in label_to_instance:
                    found_unique_label = True
                else:
                    i+=1


            if type(node) is tpot2.GraphPipeline:
                label_remapping[node] = label
                label_to_instance[label] = node
            else:
                label_remapping[node] = label
                label_to_instance[label] = node.method_class(**node.hyperparameters)

        estimator_graph = nx.relabel_nodes(estimator_graph, label_remapping)

        for label, instance in label_to_instance.items():
            estimator_graph.nodes[label]["instance"] = instance

        return tpot2.GraphPipeline(graph=estimator_graph, **graph_pipeline_args)

    def export_pipeline(self, **graph_pipeline_args):
        estimator_graph = self.flatten_pipeline()

        #mapping = {node:node.method_class(**node.hyperparameters) for node in estimator_graph}
        label_remapping = {}
        label_to_instance = {}

        for node in estimator_graph:
            found_unique_label = False
            i=1
            while not found_unique_label:
                label = "{0}_{1}".format(node.method_class.__name__, i)
                if label not in label_to_instance:
                    found_unique_label = True
                else:
                    i+=1

            label_remapping[node] = label
            label_to_instance[label] = node.method_class(**node.hyperparameters)

        estimator_graph = nx.relabel_nodes(estimator_graph, label_remapping)

        for label, instance in label_to_instance.items():
            estimator_graph.nodes[label]["instance"] = instance

        return tpot2.GraphPipeline(graph=estimator_graph, **graph_pipeline_args)

    def export_baikal(self,):
        graph = self.flatten_pipeline()
        toposorted = list(nx.topological_sort(graph))
        toposorted.reverse()
        node_outputs = {}

        X = baikal.Input('X')
        y = baikal.Input('Target')

        for i in range(len(toposorted)):
            node = toposorted[i]
            if len(list(graph.successors(node))) == 0: #If this node had no inputs use X
                this_inputs = X
            else: #in node has inputs, get those
                this_inputs = [node_outputs[child] for child in graph.successors(node)]

            this_output = baikal.make_step(node.method_class, class_name=node.method_class.__name__)(**node.hyperparameters)(this_inputs,y)
            node_outputs[node] = this_output

            if i == len(toposorted)-1: #last method doesn't need transformed.
                return baikal.Model(inputs=X, outputs=this_output, targets=y)


    def plot(self):
        G = self.flatten_pipeline().reverse() #self.graph.reverse()
        #TODO clean this up
        try:
            pos = nx.planar_layout(G)  # positions for all nodes
        except:
            pos = nx.shell_layout(G)
        # nodes
        options = {'edgecolors': 'tab:gray', 'node_size': 800, 'alpha': 0.9}
        nodelist = list(G.nodes)
        node_color = [plt.cm.Set1(G.nodes[n]['recursive depth']) for n in G]

        fig, ax = plt.subplots()

        nx.draw(G, pos, nodelist=nodelist, node_color=node_color, ax=ax,  **options)


        '''edgelist = []
        for n in n1.node_set:
            for child in n.children:
                edgelist.append((n,child))'''

        # edges
        #nx.draw_networkx_edges(G, pos, width=3.0, arrows=True)
        '''nx.draw_networkx_edges(
            G,
            pos,
            edgelist=[edgelist],
            width=8,
            alpha=0.5,
            edge_color='tab:red',
        )'''



        # some math labels
        labels = {}
        for i, n in enumerate(G.nodes):
            labels[n] = n.method_class.__name__ + "\n" + str(n.hyperparameters)


        nx.draw_networkx_labels(G, pos, labels,ax=ax, font_size=7, font_color='black')

        plt.tight_layout()
        plt.axis('off')
        plt.show()


    #############

    #TODO currently does not correctly return false when adding a leaf causes a duplicate node that is later merged
    def mutate(self, rng_=None):
        rng = np.random.default_rng(rng_)
        self.key = None
        graph = self.select_graphindividual(rng_=rng)
        return graph._mutate(rng_=rng)

    def _mutate(self, rng_=None):
        rng = np.random.default_rng(rng_)
        rng.shuffle(self.mutate_methods_list)
        for mutate_method in self.mutate_methods_list:
            if mutate_method(rng_=rng):
                self._merge_duplicated_nodes()

                if self.__debug:
                    print(mutate_method)

                    if self.root not in self.graph.nodes:
                        print('lost root something went wrong with ', mutate_method)

                    if len(self.graph.predecessors(self.root)) > 0:
                        print('root has parents ', mutate_method)

                    if any([n in nx.ancestors(self.graph,n) for n in self.graph.nodes]):
                        print('a node is connecting to itself...')

                    if self.__debug:
                        try:
                            nx.find_cycle(self.graph)
                            print('something went wrong with ', mutate_method)
                        except:
                            pass

                return True

        return False

    def _mutate_row_subsets(self, rng_=None):
        rng = np.random.default_rng(rng_)
        if self.unique_subset_values is not None:
            self.row_subset_selector.mutate(rng_=rng)


    def _mutate_hyperparameters(self, rng_=None):
        '''
        Mutates the hyperparameters for a randomly chosen node in the graph.
        '''
        rng = np.random.default_rng(rng_)
        sorted_nodes_list = list(self.graph.nodes)
        rng.shuffle(sorted_nodes_list)
        completed_one = False
        for node in sorted_nodes_list:
            if isinstance(node,GraphIndividual):
                continue
            if isinstance(self.select_config_dict(node)[node.method_class], dict):
                continue

            if not completed_one:
                _,_, completed_one = get_hyperparameter(self.select_config_dict(node)[node.method_class], rng_=rng, nodelabel=node,  alpha=self.hyperparameter_alpha, hyperparameter_probability=self.hyperparameter_probability)
            else:
                if self.hyper_node_probability > rng.random():
                    get_hyperparameter(self.select_config_dict(node)[node.method_class], rng_=rng, nodelabel=node,  alpha=self.hyperparameter_alpha, hyperparameter_probability=self.hyperparameter_probability)

        return completed_one




    def _mutate_replace_node(self, rng_=None):
        '''
        Replaces the method in a randomly chosen node by a method from the available methods for that node.

        '''
        rng = np.random.default_rng(rng_)
        sorted_nodes_list = list(self.graph.nodes)
        rng.shuffle(sorted_nodes_list)
        for node in sorted_nodes_list:
            if isinstance(node,GraphIndividual):
                continue
            new_node = create_node(self.select_config_dict(node), rng_=rng)
            #check if new node and old node are the same
            #TODO: add attempts?
            if node.method_class != new_node.method_class or node.hyperparameters != new_node.hyperparameters:
                nx.relabel_nodes(self.graph, {new_node:node}, copy=False)
                return True

        return False


    def _mutate_remove_node(self, rng_=None):
        '''
        Removes a randomly chosen node and connects its parents to its children.
        If the node is the only leaf for an inner node and 'leaf_config_dict' is not none, we do not remove it.
        '''
        rng = np.random.default_rng(rng_)
        nodes_list = list(self.graph.nodes)
        nodes_list.remove(self.root)
        leaves = graph_utils.get_leaves(self.graph)

        while len(nodes_list) > 0:
            node = rng.choice(nodes_list)
            nodes_list.remove(node)

            if self.leaf_config_dict is not None and len(list(nx.descendants(self.graph,node))) == 0 : #if the node is a leaf
                if len(leaves) <= 1:
                    continue #dont remove the last leaf
                leaf_parents = self.graph.predecessors(node)

                # if any of the parents of the node has one one child, continue
                if any([len(list(self.graph.successors(lp))) < 2 for lp in leaf_parents]): #dont remove a leaf if it is the only input into another node.
                    continue

                graph_utils.remove_and_stitch(self.graph, node)
                graph_utils.remove_nodes_disconnected_from_node(self.graph, self.root)
                return True

            else:
                graph_utils.remove_and_stitch(self.graph, node)
                graph_utils.remove_nodes_disconnected_from_node(self.graph, self.root)
                return True

        return False

    def _mutate_remove_edge(self, rng_=None):
        '''
        Deletes an edge as long as deleting that edge does not make the graph disconnected.
        '''
        rng = np.random.default_rng(rng_)
        sorted_nodes_list = list(self.graph.nodes)
        rng.shuffle(sorted_nodes_list)
        for child_node in sorted_nodes_list:
            parents = list(self.graph.predecessors(child_node))
            if len(parents) > 1: # if it has more than one parent, you can remove an edge (if this is the only child of a node, it will become a leaf)

                for parent_node in parents:
                    # if removing the egde will make the parent_node a leaf node, skip
                    if self.leaf_config_dict is not None and len(list(self.graph.successors(parent_node))) < 2:
                        continue

                    self.graph.remove_edge(parent_node, child_node)
                    return True
        return False

    def _mutate_add_edge(self, rng_=None):
        '''
        Randomly add an edge from a node to another node that is not an ancestor of the first node.
        '''
        rng = np.random.default_rng(rng_)
        sorted_nodes_list = list(self.graph.nodes)
        rng.shuffle(sorted_nodes_list)
        for child_node in sorted_nodes_list:
            for parent_node in sorted_nodes_list:
                if self.leaf_config_dict is not None:
                    if len(list(self.graph.successors(parent_node))) == 0:
                        continue

                # skip if
                # - parent and child are the same node
                # - edge already exists
                # - child is an ancestor of parent
                if  (child_node is not parent_node) and not self.graph.has_edge(parent_node,child_node) and (child_node not in nx.ancestors(self.graph, parent_node)):
                    self.graph.add_edge(parent_node,child_node)
                    return True

        return False


    def _mutate_insert_leaf(self, rng_=None):
        rng = np.random.default_rng(rng_)
        if self.max_size > self.graph.number_of_nodes():
            sorted_nodes_list = list(self.graph.nodes)
            rng.shuffle(sorted_nodes_list) #TODO: sort by number of children and/or parents? bias model one way or another
            for node in sorted_nodes_list:
                #if leafs are protected, check if node is a leaf
                #if node is a leaf, skip because we don't want to add node on top of node
                if (self.leaf_config_dict is not None #if leafs are protected
                    and   len(list(self.graph.successors(node))) == 0 #if node is leaf
                    and  len(list(self.graph.predecessors(node))) > 0 #except if node is root, in which case we want to add a leaf even if it happens to be a leaf too
                    ):

                    continue

                #If node *is* the root or is not a leaf, add leaf node. (dont want to add leaf on top of leaf)
                if self.leaf_config_dict is not None:
                    new_node = create_node(self.leaf_config_dict, rng_=rng)
                else:
                    new_node = create_node(self.inner_config_dict, rng_=rng)

                self.graph.add_node(new_node)
                self.graph.add_edge(node, new_node)
                return True

        return False

    def _mutate_insert_bypass_node(self, rng_=None):
        rng = np.random.default_rng(rng_)
        if self.max_size > self.graph.number_of_nodes():
            sorted_nodes_list = list(self.graph.nodes)
            sorted_nodes_list2 = list(self.graph.nodes)
            rng.shuffle(sorted_nodes_list) #TODO: sort by number of children and/or parents? bias model one way or another
            rng.shuffle(sorted_nodes_list2)
            for node in sorted_nodes_list:
                for child_node in sorted_nodes_list2:
                    if child_node is not node and child_node not in nx.ancestors(self.graph, node):
                        if self.leaf_config_dict is not None:
                            #If if we are protecting leafs, dont add connection into a leaf
                            if len(list(nx.descendants(self.graph,node))) ==0 :
                                continue

                            new_node = create_node(config_dict = self.inner_config_dict, rng_=rng)

                            self.graph.add_node(new_node)
                            self.graph.add_edges_from([(node, new_node), (new_node, child_node)])
                            return True

        return False


    def _mutate_insert_inner_node(self, rng_=None):
        rng = np.random.default_rng(rng_)
        if self.max_size > self.graph.number_of_nodes():
            sorted_nodes_list = list(self.graph.nodes)
            sorted_nodes_list2 = list(self.graph.nodes)
            rng.shuffle(sorted_nodes_list) #TODO: sort by number of children and/or parents? bias model one way or another
            rng.shuffle(sorted_nodes_list2)
            for node in sorted_nodes_list:
                #loop through children of node
                for child_node in list(self.graph.successors(node)):

                    if child_node is not node and child_node not in nx.ancestors(self.graph, node):
                        if self.leaf_config_dict is not None:
                            #If if we are protecting leafs, dont add connection into a leaf
                            if len(list(nx.descendants(self.graph,node))) ==0 :
                                continue

                            new_node = create_node(config_dict = self.inner_config_dict, rng_=rng)

                            self.graph.add_node(new_node)
                            self.graph.add_edges_from([(node, new_node), (new_node, child_node)])
                            self.graph.remove_edge(node, child_node)
                            return True

        return False

    ######################################################
    # Crossover

    def get_graphs(self):
        graphs = [self]
        self.graph.graph['depth'] = 0
        self.graph.graph['recursive depth'] = 0
        for node in self.graph.nodes:
            if isinstance(node, GraphIndividual):
                node.graph.graph['depth'] = nx.shortest_path_length(self.graph, self.root, node)
                graphs = graphs + node._get_graphs(depth=1)

        return graphs


    def _get_graphs(self, depth=1):
        graphs = [self]
        self.graph.graph['recursive depth'] = depth
        for node in self.graph.nodes:
            if isinstance(node, GraphIndividual):
                node.graph.graph['depth'] = nx.shortest_path_length(self.graph, self.root, node)
                graphs = graphs + node._get_graphs(depth=depth+1)

        return graphs


    def select_graphindividual(self, rng_=None):
        rng = np.random.default_rng(rng_)
        graphs = self.get_graphs()
        weights = [g.graph.number_of_nodes() for g in graphs]
        w_sum = sum(weights)
        weights = [w / w_sum for w in weights] # generate probabilities based on sum of weights
        return rng.choice(graphs, p=weights)


    def select_graph_same_recursive_depth(self,ind1,ind2,rng_=None):
        rng = np.random.default_rng(rng_)

        graphs1 = ind1.get_graphs()
        weights1 = [g.graph.number_of_nodes() for g in graphs1]
        w1_sum = sum(weights1)
        weights1 = [w / w1_sum for w in weights1]

        graphs2 = ind2.get_graphs()
        weights2 = [g.graph.number_of_nodes() for g in graphs2]
        w2_sum = sum(weights2)
        weights2 = [w / w2_sum for w in weights2]

        g1_sorted_graphs = random_weighted_sort(graphs1, weights1, rng)
        g2_sorted_graphs = random_weighted_sort(graphs2, weights2, rng)

        for g1, g2 in zip(g1_sorted_graphs, g2_sorted_graphs):
            if g1.graph.graph['depth'] == g2.graph.graph['depth'] and g1.graph.graph['recursive depth'] == g2.graph.graph['recursive depth']:
                return g1, g2

        return ind1,ind2

    def crossover(self, ind2, rng_=None):
        '''
        self is the first individual, ind2 is the second individual
        If crossover_same_depth, it will select graphindividuals at the same recursive depth.
        Otherwise, it will select graphindividuals randomly from the entire graph and its subgraphs.

        This does not impact graphs without subgraphs. And it does not impacts nodes that are not graphindividuals. Cros
        '''

        rng = np.random.default_rng(rng_)

        self.key = None
        ind2.key = None
        if self.crossover_same_recursive_depth:
            # selects graphs from the same recursive depth and same depth from the root
            g1, g2 = self.select_graph_same_recursive_depth(self, ind2, rng_=rng)


        else:
            g1 = self.select_graphindividual(rng_=rng)
            g2 = ind2.select_graphindividual(rng_=rng)

        return g1._crossover(g2, rng_=rng)

    def _crossover(self, Graph, rng_=None):
        rng = np.random.default_rng(rng_)

        rng.shuffle(self.crossover_methods_list)
        for crossover_method in self.crossover_methods_list:
            if crossover_method(Graph, rng_=rng):
                self._merge_duplicated_nodes()
                return True

        if self.__debug:
            try:
                nx.find_cycle(self.graph)
                print('something went wrong with ', crossover_method)
            except:
                pass

        return False


    def _crossover_row_subsets(self, G2, rng_=None):
        rng = np.random.default_rng(rng_)
        if self.unique_subset_values is not None and G2.unique_subset_values is not None:
            self.row_subset_selector.crossover(G2.row_subset_selector, rng_=rng)


    def _crossover_swap_node(self, G2, rng_=None):
        '''
        Swaps randomly chosen node from Parent1 with a randomly chosen node from Parent2.
        '''
        rng = np.random.default_rng(rng_)

        if self.crossover_same_depth:
            pair_gen = graph_utils.select_nodes_same_depth(self.graph, self.root, G2.graph, G2.root, rng_=rng)
        else:
            pair_gen = graph_utils.select_nodes_randomly(self.graph, G2.graph, rng_=rng)

        for node1, node2 in pair_gen:
            if not (node1 is self.root or node2 is G2.root): #TODO: allow root

                n1_s = self.graph.successors(node1)
                n1_p = self.graph.predecessors(node1)

                n2_s = G2.graph.successors(node2)
                n2_p = G2.graph.predecessors(node2)

                self.graph.remove_node(node1)
                G2.graph.remove_node(node2)

                self.graph.add_node(node2)

                self.graph.add_edges_from([ (node2, n) for n in n1_s])
                G2.graph.add_edges_from([ (node1, n) for n in n2_s])

                self.graph.add_edges_from([ (n, node2) for n in n1_p])
                G2.graph.add_edges_from([ (n, node1) for n in n2_p])

                return True
        return False



    def _crossover_swap_branch(self, G2, rng_=None):
        '''
        swaps a branch from parent1 with a branch from parent2. does not modify parent2
        '''
        rng = np.random.default_rng(rng_)

        if self.crossover_same_depth:
            pair_gen = graph_utils.select_nodes_same_depth(self.graph, self.root, G2.graph, G2.root, rng_=rng)
        else:
            pair_gen = graph_utils.select_nodes_randomly(self.graph, G2.graph, rng_=rng)

        for node1, node2 in pair_gen:
            #TODO: if root is in inner_config_dict, then do use it?
            if node1 is self.root or node2 is G2.root: #dont want to add root as inner node
                continue

            #check if node1 is a leaf and leafs are protected, don't add an input to the leave
            if self.leaf_config_dict is not None: #if we are protecting leaves,
                node1_is_leaf = len(list(self.graph.successors(node1))) == 0
                node2_is_leaf = len(list(G2.graph.successors(node2))) == 0
                #if not ((node1_is_leaf and node1_is_leaf) or (not node1_is_leaf and not node2_is_leaf)): #if node1 is a leaf
                if (node1_is_leaf and (not node2_is_leaf)) or ( (not node1_is_leaf) and node2_is_leaf):
                    #only continue if node1 and node2 are both leaves or both not leaves
                    continue

            temp_graph_1 = self.graph.copy()
            temp_graph_1.remove_node(node1)
            graph_utils.remove_nodes_disconnected_from_node(temp_graph_1, self.root)

            #isolating the branch
            branch2 = G2.graph.copy()
            n2_descendants = nx.descendants(branch2,node2)
            for n in list(branch2.nodes):
                if n not in n2_descendants and n is not node2: #removes all nodes not in the branch
                    branch2.remove_node(n)

            branch2 = copy.deepcopy(branch2)
            branch2_root = graph_utils.get_roots(branch2)[0]
            temp_graph_1.add_edges_from(branch2.edges)
            for p in list(self.graph.predecessors(node1)):
                temp_graph_1.add_edge(p,branch2_root)

            if temp_graph_1.number_of_nodes() > self.max_size:
                continue

            self.graph = temp_graph_1

            return True
        return False

    #TODO: Currently returns true even if hyperparameters are blank
    def _crossover_hyperparameters(self, G2, rng_=None):
        '''
        Swaps the hyperparamters of one randomly chosen node in Parent1 with the hyperparameters of randnomly chosen node in Parent2.
        '''
        rng = np.random.default_rng(rng_)

        if self.crossover_same_depth:
            pair_gen = graph_utils.select_nodes_same_depth(self.graph, self.root, G2.graph, G2.root, rng_=rng)
        else:
            pair_gen = graph_utils.select_nodes_randomly(self.graph, G2.graph, rng_=rng)

        for node1, node2 in pair_gen:
            if isinstance(node1,GraphIndividual) or isinstance(node2,GraphIndividual):
                continue

            if node1.method_class == node2.method_class:
                tmp = node1.hyperparameters
                node1.hyperparameters = node2.hyperparameters
                node2.hyperparameters = tmp
                return True

        return False

    #not including the nodes, just their children
    #Finds leaves attached to nodes and swaps them
    def _crossover_swap_leaf_at_node(self, G2, rng_=None):
        rng = np.random.default_rng(rng_)

        if self.crossover_same_depth:
            pair_gen = graph_utils.select_nodes_same_depth(self.graph, self.root, G2.graph, G2.root, rng_=rng)
        else:
            pair_gen = graph_utils.select_nodes_randomly(self.graph, G2.graph, rng_=rng)

        success = False
        for node1, node2 in pair_gen:
            # if leaves are protected node1 and node2 must both be leaves or both be inner nodes
            if self.leaf_config_dict is not None and not (len(list(self.graph.successors(node1)))==0 ^ len(list(G2.graph.successors(node2)))==0):
                continue
            #self_leafs = [c for c in nx.descendants(self.graph,node1) if len(list(self.graph.successors(c)))==0 and c is not node1]
            node_leafs = [c for c in nx.descendants(G2.graph,node2) if len(list(G2.graph.successors(c)))==0 and c is not node2]

            # if len(self_leafs) >0:
            #     for c in self_leafs:
            #         if random.choice([True,False]):
            #             self.graph.remove_node(c)
            #             G2.graph.add_edge(node2, c)
            #             success = True

            if len(node_leafs) >0:
                for c in node_leafs:
                    if rng.choice([True,False]):
                        G2.graph.remove_node(c)
                        self.graph.add_edge(node1, c)
                        success = True

        return success


    def _crossover_take_branch(self, G2, rng_=None):
        '''
        Takes a subgraph from Parent2 and add it to a randomly chosen node in Parent1.
        '''
        rng = np.random.default_rng(rng_)

        if self.crossover_same_depth:
            pair_gen = graph_utils.select_nodes_same_depth(self.graph, self.root, G2.graph, G2.root, rng_=rng)
        else:
            pair_gen = graph_utils.select_nodes_randomly(self.graph, G2.graph, rng_=rng)

        for node1, node2 in pair_gen:
            #TODO: if root is in inner_config_dict, then do use it?
            if node2 is G2.root: #dont want to add root as inner node
                continue


            #check if node1 is a leaf and leafs are protected, don't add an input to the leave
            if self.leaf_config_dict is not None and len(list(self.graph.successors(node1))) == 0:
                continue

            #icheck if node2 is graph individual
            # if isinstance(node2,GraphIndividual):
            #     if not ((isinstance(node2,GraphIndividual) and ("Recursive" in self.inner_config_dict or "Recursive" in self.leaf_config_dict))):
            #         continue

            #isolating the branch
            branch2 = G2.graph.copy()
            n2_descendants = nx.descendants(branch2,node2)
            for n in list(branch2.nodes):
                if n not in n2_descendants and n is not node2: #removes all nodes not in the branch
                    branch2.remove_node(n)

            #if node1 plus node2 branch has more than max_children, skip
            if branch2.number_of_nodes() + self.graph.number_of_nodes() > self.max_size:
                continue

            branch2 = copy.deepcopy(branch2)
            branch2_root = graph_utils.get_roots(branch2)[0]
            self.graph.add_edges_from(branch2.edges)
            self.graph.add_edge(node1,branch2_root)

            return True
        return False

    #TODO: swap all leaf nodes
    def _crossover_swap_all_leafs(self, G2, rng_=None):
        pass


    #TODO: currently ignores ensembles, make it include nodes inside of ensembles
    def optimize(self, rng_, objective_function, steps=5):
        rng = np.random.default_rng(rng_)
        rng.shuffle(self.optimize_methods_list) #select an optimization method
        for optimize_method in self.optimize_methods_list:
            if optimize_method(rng, objective_function, steps=steps):
                return True

    #optimize the hyperparameters of one method to improve the entire pipeline
    def _optimize_optuna_single_method_full_pipeline(self, rng_, objective_function, steps=5):
        rng = np.random.default_rng(rng_)
        nodes_list = list(self.graph.nodes)
        rng.shuffle(nodes_list) #TODO: sort by number of children and/or parents? bias model one way or another
        for node in nodes_list:
            if not isinstance(node, NodeLabel) or isinstance(self.select_config_dict(node)[node.method_class],dict):
                continue
            else:
                study = optuna.create_study()

                def objective(trial):
                    params = self.select_config_dict(node)[node.method_class](trial)
                    node.hyperparameters = params

                    trial.set_user_attr('params', params)
                    try:
                        return objective_function(self)
                    except:
                        return np.NAN

                study.optimize(objective, n_trials=steps)
                node.hyperparameters = study.best_trial.user_attrs['params']
                return True


    #optimize the hyperparameters of all methods simultaneously to improve the entire pipeline
    def _optimize_optuna_all_methods_full_pipeline(self, rng_, objective_function, steps=5):
        nodes_list = list(self.graph.nodes)
        study = optuna.create_study()
        nodes_to_optimize = []
        for node in nodes_list:
            if not isinstance(node, NodeLabel) or isinstance(self.select_config_dict(node)[node.method_class],dict):
                continue
            else:
                nodes_to_optimize.append(node)

        def objective(trial):
            param_list = []
            for i, node in enumerate(nodes_to_optimize):
                params = self.select_config_dict(node)[node.method_class](trial, name=f'node_{i}')
                node.hyperparameters = params
                param_list.append(params)

            trial.set_user_attr('params', param_list)

            try:
                return objective_function(self)
            except:
                return np.NAN

        study.optimize(objective, n_trials=steps)
        best_params = study.best_trial.user_attrs['params']

        for node, params in zip(nodes_to_optimize,best_params):
            node.hyperparameters = params

        return True


    def _cached_transform(cache_nunber=0):
        #use a cache for models at each CV fold?
        #cache just transformations at each fold?
        #TODO how to separate full model?
        pass

    def __str__(self):
        try:
            return f"<GraphIdnividual {0}".format(self.export_pipeline().__str__())
        except:
            return "<Invalid GraphIdnividual>"

    def unique_id(self) -> GraphKey:
        if self.key is None:
            g = self.flatten_pipeline()
            for n in g.nodes:
                if "subset_values" in g.nodes[n]:
                    g.nodes[n]['label'] = {n.method_class: n.hyperparameters, "subset_values":g.nodes[n]["subset_values"]}
                else:
                    g.nodes[n]['label'] = {n.method_class: n.hyperparameters}

                g.nodes[n]['method_class'] = n.method_class #TODO making this transformation doesn't feel very clean?
                g.nodes[n]['hyperparameters'] = n.hyperparameters

            g = nx.convert_node_labels_to_integers(g)
            self.key = GraphKey(graph=g)

        return self.key

    def full_node_list(self):
        node_list = list(self.graph.nodes)
        for node in node_list:
            if isinstance(node, GraphIndividual):
                node_list.pop(node_list.index(node))
                node_list.extend(node.graph.nodes)
        return node_list




def create_node(config_dict, rng_=None):
    '''
    Takes a config_dict and returns a node with a random method_class and hyperparameters
    '''
    rng = np.random.default_rng(rng_)
    method_class = rng.choice(list(config_dict.keys()))
    #if method_class == GraphIndividual or method_class == 'Recursive':
    if method_class == 'Recursive':
        node = GraphIndividual(**config_dict[method_class])
    else:
        hyperparameters, params, _ = get_hyperparameter(config_dict[method_class], rng_=rng, nodelabel=None)

        node = NodeLabel(
                                        method_class=method_class,
                                        hyperparameters=hyperparameters
                                        )
        node._params = params

    return node


def random_weighted_sort(l,weights, rng_=None):
    rng = np.random.default_rng(rng_)
    sorted_l = []
    indeces = {i: weights[i] for i in range(len(l))}
    while len(indeces) > 0:
        keys = list(indeces.keys())
        p = np.array([indeces[k] for k in keys])
        p = p / p.sum()
        next_item = rng.choice(list(indeces.keys()), p=p)
        indeces.pop(next_item)
        sorted_l.append(l[next_item])

    return sorted_l


def get_hyperparameter(config_func, rng_, nodelabel=None,  alpha=1, hyperparameter_probability=1):
    rng = np.random.default_rng(rng_)
    changed = False
    if isinstance(config_func, dict):
        return config_func, None, changed

    if nodelabel is not None:
        trial = config.hyperparametersuggestor.Trial(rng_=rng, old_params=nodelabel._params, alpha=alpha, hyperparameter_probability=hyperparameter_probability)
        new_params = config_func(trial)
        changed = trial._params != nodelabel._params
        nodelabel._params = trial._params
        nodelabel.hyperparameters = new_params
    else:
        trial = config.hyperparametersuggestor.Trial(rng_=rng, old_params=None, alpha=alpha, hyperparameter_probability=hyperparameter_probability)
        new_params = config_func(trial)

    return  new_params, trial._params, changed


import numpy as np
import sklearn
import sklearn.base
import tpot2
import pandas as pd

from .cross_val_utils import cross_val_score_objective

def convert_parents_tuples_to_integers(row, object_to_int):
    if type(row) == list or type(row) == np.ndarray or type(row) == tuple:
        return tuple(object_to_int[obj] for obj in row)
    else:
        return np.nan

def apply_make_pipeline(graphindividual, preprocessing_pipeline=None):
    try:
        if preprocessing_pipeline is None:
            return graphindividual.export_pipeline()
        else:
            return sklearn.pipeline.make_pipeline(sklearn.base.clone(preprocessing_pipeline), graphindividual.export_pipeline())
    except:
        return None

def get_configuration_dictionary(options, n_samples, n_features, classification, random_state=None, cv=None, subsets=None, feature_names=None, n_classes=None):
    if options is None:
        return options

    if isinstance(options, dict):
        return recursive_with_defaults(options, n_samples, n_features, classification, random_state=None, cv=None, subsets=subsets, feature_names=feature_names, n_classes=n_classes)

    if not isinstance(options, list):
        options = [options]

    config_dict = {}

    for option in options:

        if option == "selectors":
            config_dict.update(tpot2.config.make_selector_config_dictionary(random_state=random_state, classifier=classification))

        elif option == "classifiers":
            config_dict.update(tpot2.config.make_classifier_config_dictionary(random_state=random_state, n_samples=n_samples, n_classes=n_classes))

        elif option == "classifiers_sklearnex":
            config_dict.update(tpot2.config.make_sklearnex_classifier_config_dictionary(random_state=random_state, n_samples=n_samples, n_classes=n_classes))

        elif option == "regressors":
            config_dict.update(tpot2.config.make_regressor_config_dictionary(random_state=random_state, cv=cv, n_samples=n_samples))

        elif option == "regressors_sklearnex":
            config_dict.update(tpot2.config.make_sklearnex_regressor_config_dictionary(random_state=random_state, n_samples=n_samples))

        elif option == "transformers":
            config_dict.update(tpot2.config.make_transformer_config_dictionary(random_state=random_state, n_features=n_features))

        elif option == "arithmetic_transformer":
            config_dict.update(tpot2.config.make_arithmetic_transformer_config_dictionary())

        elif option == "feature_set_selector":
            config_dict.update(tpot2.config.make_FSS_config_dictionary(subsets, n_features, feature_names=feature_names))

        elif option == "skrebate":
            config_dict.update(tpot2.config.make_skrebate_config_dictionary(n_features=n_features))

        elif option == "MDR":
            config_dict.update(tpot2.config.make_MDR_config_dictionary())

        elif option == "continuousMDR":
            config_dict.update(tpot2.config.make_ContinuousMDR_config_dictionary())

        elif option == "FeatureEncodingFrequencySelector":
            config_dict.update(tpot2.config.make_FeatureEncodingFrequencySelector_config_dictionary())

        elif option == "genetic encoders":
            config_dict.update(tpot2.config.make_genetic_encoders_config_dictionary())

        elif option == "passthrough":
            config_dict.update(tpot2.config.make_passthrough_config_dictionary())


        else:
            config_dict.update(recursive_with_defaults(option, n_samples, n_features, classification, random_state, cv, subsets=subsets, feature_names=feature_names, n_classes=n_classes))

    if len(config_dict) == 0:
        raise ValueError("No valid configuration options were provided. Please check the options you provided and try again.")

    return config_dict

def recursive_with_defaults(config_dict, n_samples, n_features, classification, random_state=None, cv=None, subsets=None, feature_names=None, n_classes=None):

    for key in 'leaf_config_dict', 'root_config_dict', 'inner_config_dict', 'Recursive':
        if key in config_dict:
            value = config_dict[key]
            if key=="Resursive":
                config_dict[key] = recursive_with_defaults(value, n_samples, n_features, classification, random_state, cv, subsets=None, feature_names=None, n_classes=None)
            else:
                config_dict[key] = get_configuration_dictionary(value, n_samples, n_features, classification, random_state, cv, subsets, feature_names, n_classes)

    return config_dict



def objective_function_generator(pipeline, x,y, scorers, cv, other_objective_functions, memory=None, cross_val_predict_cv=None, subset_column=None, step=None, budget=None, generation=1,is_classification=True):
    pipeline = pipeline.export_pipeline(memory=memory, cross_val_predict_cv=cross_val_predict_cv, subset_column=subset_column)
    if budget is not None and budget < 1:
        if is_classification:
            x,y = sklearn.utils.resample(x,y, stratify=y, n_samples=int(budget*len(x)), replace=False, random_state=1)
        else:
            x,y = sklearn.utils.resample(x,y, n_samples=int(budget*len(x)), replace=False, random_state=1)

        if isinstance(cv, int) or isinstance(cv, float):
            n_splits = cv
        else:
            n_splits = cv.n_splits

    if len(scorers) > 0:
        cv_obj_scores = cross_val_score_objective(sklearn.base.clone(pipeline),x,y,scorers=scorers, cv=cv , fold=step)
    else:
        cv_obj_scores = []

    if other_objective_functions is not None and len(other_objective_functions) >0:
        other_scores = [obj(sklearn.base.clone(pipeline)) for obj in other_objective_functions]
        #flatten
        other_scores = np.array(other_scores).flatten().tolist()
    else:
        other_scores = []

    return np.concatenate([cv_obj_scores,other_scores])

def val_objective_function_generator(pipeline, X_train, y_train, X_test, y_test, scorers, other_objective_functions, memory, cross_val_predict_cv, subset_column):
    #subsample the data
    pipeline = pipeline.export_pipeline(memory=memory, cross_val_predict_cv=cross_val_predict_cv, subset_column=subset_column)
    fitted_pipeline = sklearn.base.clone(pipeline)
    fitted_pipeline.fit(X_train, y_train)

    if len(scorers) > 0:
        scores =[sklearn.metrics.get_scorer(scorer)(fitted_pipeline, X_test, y_test) for scorer in scorers]

    other_scores = []
    if other_objective_functions is not None and len(other_objective_functions) >0:
        other_scores = [obj(sklearn.base.clone(pipeline)) for obj in other_objective_functions]

    return np.concatenate([scores,other_scores])


def remove_underrepresented_classes(x, y, min_count):
    if isinstance(y, (np.ndarray, pd.Series)):
        unique, counts = np.unique(y, return_counts=True)
        if min(counts) >= min_count:
            return x, y
        keep_classes = unique[counts >= min_count]
        mask = np.isin(y, keep_classes)
        x = x[mask]
        y = y[mask]
    elif isinstance(y, pd.DataFrame):
        counts = y.apply(pd.Series.value_counts)
        if min(counts) >= min_count:
            return x, y
        keep_classes = counts.index[counts >= min_count].tolist()
        mask = y.isin(keep_classes).all(axis=1)
        x = x[mask]
        y = y[mask]
    else:
        raise TypeError("y must be a numpy array or a pandas Series/DataFrame")
    return x, y


def convert_to_float(x):
    try:
        return float(x)
    except ValueError:
        return x




def check_if_y_is_encoded(y):
    '''
    checks if the target y is composed of sequential ints from 0 to N
    '''
    y = sorted(set(y))
    return all(i == j for i, j in enumerate(y))



from .average_path_length import average_path_length_objective
from .number_of_nodes import number_of_nodes_objective
from .number_of_leaves import number_of_leaves_scorer, number_of_leaves_objective
from .complexity import complexity_scorer


#these scorers are calculated per fold of CV on the fitted pipeline for that fold
SCORERS =       {   
                    "complexity_scorer": complexity_scorer
                }

#these objectives are calculated once on unfitted models as secondary objectives
OBJECTIVES =    {  "average_path_length_objective": average_path_length_objective,
                    "number_of_nodes_objective": number_of_nodes_objective,
                    "number_of_leaves_objective": number_of_leaves_objective
                }


from tpot2.individual_representations.graph_pipeline_individual.individual import *
import optuna
import numpy as np
import copy
import dask
import traceback
import functools

# labels all nodes in the graph with a unique ID.
# This allows use to identify exact nodes in the copies on the graph.
# This is necessary since copies of the graph use different NodeLabel object instances as keys, making it hard to identify which are the same nodes.
def label_nodes_in_graphindividual(graphindividual):
    nodes_list = graphindividual.full_node_list()
    for i, node in enumerate(nodes_list):
        if not isinstance(node, NodeLabel):
            continue
        else:
            node.label = f'node_{i}'


def optuna_optimize_full_graph(graphindividual, objective_function, objective_function_weights, steps=5, relabel=True, verbose=0, max_eval_time_seconds=60*5, max_time_seconds=60*20, n_returned_models='all', study=None, **objective_kwargs):
    if relabel:
        label_nodes_in_graphindividual(graphindividual)
    
    graphindividual = copy.deepcopy(graphindividual)
    nodes_list = graphindividual.full_node_list()


    nodes_to_optimize = []
    for node in nodes_list:
        if not isinstance(node, NodeLabel) or isinstance(graphindividual.select_config_dict(node)[node.method_class],dict):
            continue
        else:
            nodes_to_optimize.append(node)

    def objective(trial):
        param_dict = dict()
        graphindividual.key = None
        for node in nodes_to_optimize:
            params = graphindividual.select_config_dict(node)[node.method_class](trial, name=node.label)
            node.hyperparameters = params
            param_dict[node.label] = params
        
        trial.set_user_attr('params', param_dict)
        
        try:
            scores = tpot2.objective_nan_wrapper(graphindividual, objective_function, verbose=verbose,timeout=max_eval_time_seconds,**objective_kwargs)#objective_function(graphindividual)
            trial.set_user_attr('scores', list(scores))
            if scores[0] != "INVALID" and scores[0] != "TIMEOUT":
                scores = np.array(scores) * objective_function_weights
            scores = list(scores)
            
        except Exception as e:
            print(e)
            print(traceback.format_exc())
            scores = ['INVALID']
            trial.set_user_attr('scores', scores)
        return scores

    study.optimize(objective, n_trials=steps, timeout=max_time_seconds)
   
    return study

def graph_objective(trial, graphindividual, objective_function, objective_function_weights, verbose=0, max_eval_time_seconds=60*5, **objective_kwargs):

    graphindividual = copy.deepcopy(graphindividual)
    nodes_list = graphindividual.full_node_list()


    nodes_to_optimize = []
    for node in nodes_list:
        if not isinstance(node, NodeLabel) or isinstance(graphindividual.select_config_dict(node)[node.method_class],dict):
            continue
        else:
            nodes_to_optimize.append(node)

    param_dict = dict()
    graphindividual.key = None
    for node in nodes_to_optimize:
        params = graphindividual.select_config_dict(node)[node.method_class](trial, name=node.label)
        node.hyperparameters = params
        param_dict[node.label] = params
    
    trial.set_user_attr('params', param_dict)
    
    try:
        scores = tpot2.objective_nan_wrapper(graphindividual, objective_function, verbose=verbose,timeout=max_eval_time_seconds,**objective_kwargs)#objective_function(graphindividual)
        trial.set_user_attr('scores', list(scores))
        if scores[0] != "INVALID" and scores[0] != "TIMEOUT":
            scores = np.array(scores) * objective_function_weights
        scores = list(scores)
        
    except Exception as e:
        print(e)
        print(traceback.format_exc())
        scores = ['INVALID']
        trial.set_user_attr('scores', scores)
    
    return scores


def simple_parallel_optuna(individuals,  objective_function, objective_function_weights, client, storage, steps=5, verbose=0, max_eval_time_seconds=60*5, max_time_seconds=60*20, **objective_kwargs):
    num_workers = len(client.scheduler_info()['workers'])
    worker_per_individual = max(1,int(np.floor(num_workers/len(individuals))))
    remainder = num_workers%len(individuals)

    print(len(individuals))

    directions = np.repeat('maximize',len(objective_function_weights))
    timeout = max(max_time_seconds/len(individuals), max_eval_time_seconds*2)

    studies = []
    for i, ind in enumerate(individuals):
        label_nodes_in_graphindividual(ind)
        print(ind)

        #study  = optuna.create_study(directions=directions, storage=f"{storage}", load_if_exists=False)
        backend_storage = optuna.storages.InMemoryStorage()
        study  = optuna.create_study(directions=directions, storage=backend_storage, load_if_exists=False)
        studies.append(study)

        objective = functools.partial(graph_objective, graphindividual=ind, objective_function=objective_function, objective_function_weights=objective_function_weights, verbose=verbose, max_eval_time_seconds=max_eval_time_seconds, **objective_kwargs)
        study.optimize(objective, n_trials=steps,  timeout=timeout, n_jobs=num_workers)

    all_graphs = []
    all_scores = []
    for study, ind in zip(studies,individuals):
        graphs, scores = get_all_individuals_from_study(study, ind)
        all_graphs.extend(graphs)
        all_scores.extend(scores)

    return all_graphs, all_scores




def simple_parallel_optuna_old(individuals,  objective_function, objective_function_weights, client, storage, steps=5, verbose=0, max_eval_time_seconds=60*5, max_time_seconds=60*20, **objective_kwargs):
    num_workers = len(client.scheduler_info()['workers'])
    worker_per_individual = max(1,int(np.floor(num_workers/len(individuals))))
    remainder = num_workers%len(individuals)

    print(worker_per_individual)
    print(remainder)

    directions = np.repeat('maximize',len(objective_function_weights))



    futures = []
    studies = []
    for i, ind in enumerate(individuals):
        label_nodes_in_graphindividual(ind)
        #study  = optuna.create_study(directions=directions, storage=f"{storage}", load_if_exists=False)
        backend_storage = optuna.storages.InMemoryStorage()
        dask_storage = optuna.integration.DaskStorage(storage=backend_storage, client=client)
        study  = optuna.create_study(directions=directions, storage=dask_storage, load_if_exists=False)
        studies.append(study)
        if i == 0:
            n_futures = worker_per_individual + remainder
        else:
            n_futures = worker_per_individual
        
        trials_per_thread = int(np.ceil(steps/n_futures))
        
        objective = functools.partial(graph_objective, graphindividual=ind, objective_function=objective_function, objective_function_weights=objective_function_weights, verbose=verbose, max_eval_time_seconds=max_eval_time_seconds)
        for _ in range(n_futures):
            #future = client.submit(study.optimize, objective, n_trials=trials_per_thread, pure=False, timeout=max_time_seconds,)
            future = client.submit(submit_helper, study=study, objective=objective, n_trials=trials_per_thread, timeout=max_time_seconds, pure=False, **objective_kwargs)
            futures.append(future)
            #futures.append(client.submit(optuna_optimize_full_graph, graphindividual=ind, objective_function=objective_function, objective_function_weights=objective_function_weights, steps=trials_per_thread, verbose=verbose, max_eval_time_seconds=max_eval_time_seconds, max_time_seconds=max_time_seconds, study=study, relabel=False, pure=False, **objective_kwargs))

    print(len(individuals))
    print(len(futures))
    dask.distributed.wait(futures)

    all_graphs = []
    all_scores = []
    for study, ind in zip(studies,individuals):
        graphs, scores = get_all_individuals_from_study(study, ind)
        all_graphs.extend(graphs)
        all_scores.extend(scores)

    return all_graphs, all_scores

def submit_helper(study, objective, n_trials, timeout, **kwargs):
    objective = functools.partial(objective, **kwargs)
    study.optimize(objective, n_trials=n_trials,timeout=timeout)

def get_all_individuals_from_study(study, graphindividual, n_returned_models='all'):
    all_graphs = []
    all_scores = []

    if n_returned_models == 'pareto':
        trials_list = study.best_trials
    else:
        trials_list = study.trials

    for trial in trials_list:
        if not 'scores' in trial.user_attrs:
            continue

        params = trial.user_attrs['params']
        scores = trial.user_attrs['scores']

        newgraphindividual = copy.deepcopy(graphindividual)
        newgraphindividual.key = None
        try: 
            for node in newgraphindividual.full_node_list():
                if not isinstance(node, tpot2.NodeLabel):
                    continue
                else:
                    if node.label in params:
                        node.hyperparameters = params[node.label]

            all_graphs.append(newgraphindividual)
            all_scores.append(scores)
        except Exception as e:
            print('failed to create graphindividual from trial')
            print(e)
            print(traceback.format_exc())
            print(params)
            print(newgraphindividual)
            print(newgraphindividual.graph.nodes)
            for node in newgraphindividual.full_node_list():
                print(node.label)
                

    return all_graphs, all_scores


import numpy as np

def random_selector(scores,  k, rng_=None, n_parents=1, ):
    rng = np.random.default_rng(rng_)
    chosen = rng.choice(list(range(0,len(scores))), size=k*n_parents)
    return np.reshape(chosen, (k, n_parents))


import networkx as nx
import numpy as np

def average_path_length_objective(graph_pipeline):

    path_lengths =  nx.shortest_path_length(graph_pipeline.graph, source=graph_pipeline.root)
    return np.mean(np.array(list(path_lengths.values())))+1